package org.vaadin.console;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.vaadin.console.util.ServerCommUtils;
import org.vaadin.console.util.ServerCommUtils.CallableComponent;

import com.vaadin.terminal.PaintException;
import com.vaadin.terminal.PaintTarget;
import com.vaadin.ui.AbstractComponent;
import com.vaadin.ui.Component;

/**
 * Server side component for the VTextConsole widget.
 *
 * @author Sami Ekblad / Vaadin
 *
 */
@com.vaadin.ui.ClientWidget(org.vaadin.console.client.ui.VTextConsole.class)
public class Console extends AbstractComponent implements Component.Focusable {

    private static final long serialVersionUID = 590258219352859644L;
    private String lastInput;
    private Handler handler;
    private HashMap<String, Command> commands = new HashMap<String, Command>();
    private Config config = new Config();

    private static final String DEFAULT_PS = "}> ";
    private static final String DEFAULT_GREETING = "Console ready.";
    private static final int DEFAULT_BUFFER = 0;
    private static final int DEFAULT_COLS = -1;
    private static final int DEFAULT_ROWS = -1;
    private static final boolean DEFAULT_WRAP = true;
    private static final int MAX_COLS = 500;
    private static final int MAX_ROWS = 200;

    public boolean isWrap() {
        return config.wrap;
    }

    public void setWrap(boolean wrap) {
        config.wrap = wrap;
        client.call("setWrap", wrap);
    }

    private ServerCommUtils client = new ServerCommUtils(new ClientCallback());

    /**
     * Inner class to handle client calls.
     *
     */
    public class ClientCallback implements CallableComponent {

        private static final long serialVersionUID = 3992611573500588703L;

        public void clientCalls(String method, Object[] params) {
            // TODO Auto-generated method stub

        }

        public void clientRequestedInit() {
            client.initClientWidget(config.cols, config.rows,
                    config.maxBufferSize, config.wrap, config.greeting,
                    config.ps);
        }

        public void requestRepaint() {
            Console.this.requestRepaint();
        }

    }

    /**
     * The tab order number of this field.
     */
    private int tabIndex = 0;
    @SuppressWarnings("unused")
    private Integer fontw;
    @SuppressWarnings("unused")
    private Integer fonth;
    private PrintStream printStream;
    private String lastSuggestInput;

    /**
     * An inner class for holding the configuration data.
     */
    public class Config implements Serializable {

        private static final long serialVersionUID = -812601232248504108L;

        int maxBufferSize = DEFAULT_BUFFER;
        int cols = DEFAULT_COLS;
        int rows = DEFAULT_ROWS;
        boolean wrap = DEFAULT_WRAP;
        String ps = DEFAULT_PS;
        String greeting = DEFAULT_GREETING;

    }

    public interface Handler extends Serializable {

        boolean inputReceived(Console console, String lastInput);

        void exception(Exception e, Command cmd, String[] argv);

    }

    /**
     * Commands that can be executed against the Component instance. They
     * provide convenient string to method mapping. Basically, a Command is a
     * method that that can be executed in Component. It can have parameters or
     * not.
     *
     */
    public interface Command extends Serializable {

        public Object execute(String[] argv) throws Exception;

        public String getUsage(String[] argv);
    }

    public Console(Console.Handler handler) {
        this();
        this.handler = handler;
    }

    public Console() {
        setImmediate(true);
    }

    @Override
    public void paintContent(PaintTarget target) throws PaintException {
        super.paintContent(target);
        client.paintContent(target);
    }

    @Override
    public void changeVariables(Object source, Map<String, Object> variables) {
        super.changeVariables(source, variables);
        client.changeVariables(source, variables);
        lastInput = null;
        if (variables.containsKey("width")) {
            setWidth((String) variables.get("width"));
        }
        if (variables.containsKey("height")) {
            setHeight((String) variables.get("height"));
        }
        if (variables.containsKey("fontw")) {
            fontw = (Integer) variables.get("fontw");
        }
        if (variables.containsKey("fonth")) {
            fonth = (Integer) variables.get("fonth");
        }
        if (variables.containsKey("cols")) {
            config.cols = (Integer) variables.get("cols");
        }
        if (variables.containsKey("rows")) {
            config.rows = (Integer) variables.get("rows");
        }
        if (variables.containsKey("input")) {
            lastInput = (String) variables.get("input");
            handleInput(lastInput);
        }
        if (variables.containsKey("suggest")) {
            String suggest = (String) variables.get("suggest");
            handleSuggest(suggest);
        }
    }

    protected void handleSuggest(String input) {
        boolean cancelIfNotASingleMatch = (input != null && !input
                .equals(lastSuggestInput));
        lastSuggestInput = input;
        String prefix = input.endsWith(" ") ? "" : null;
        if (prefix == null) {
            String[] argv = parseInput(input);
            if (argv != null && argv.length > 0) {
                prefix = argv[argv.length - 1];
            }
        }
        if (prefix != null) {
            List<String> matches = new ArrayList<String>();
            Set<String> cmds = getCommandNames();
            for (String cmd : cmds) {
                if (cmd.startsWith(prefix)) {
                    matches.add(cmd);
                }
            }

            // Output the original
            String prompt = input.substring(0, input.lastIndexOf(prefix));
            if (matches.size() == 0) {
                bell();
                return;
            } else if (matches.size() == 1) {
                prompt += matches.get(0) + " "; // append the single match
            } else {
                String commonPrefix = prefix;
                int maxLen = matches.get(0).length();
                for (int i = prefix.length(); i < maxLen; i++) {
                    char c = 0;
                    boolean charMatch = true;
                    for (String m : matches) {
                        if (c == 0) {
                            c = m.charAt(i);
                        } else if (i < m.length()) {
                            charMatch &= m.charAt(i) == c;
                            if (!charMatch) {
                                break;
                            }
                        } else {
                            charMatch = false;
                            break;
                        }
                    }
                    if (charMatch) {
                        commonPrefix += c;
                    }
                }
                prompt += commonPrefix;
                if (prefix.equals(commonPrefix) && !cancelIfNotASingleMatch) {
                    StringBuffer suggestions = new StringBuffer("\n");
                    for (String m : matches) {
                        suggestions.append(" " + m);
                    }
                    print(suggestions.toString());
                } else {
                    bell();
                    lastSuggestInput = prompt; // next suggest will not beep
                }

            }
            prompt(prompt);
            focus();
        }

    }

    public void bell() {
        client.call("bell");
    }

    protected Set<String> getCommandNames() {
        return new HashSet<String>(commands.keySet());
    }

    protected void handleInput(String input) {

        // Ask registered handler first
        boolean preventDefault = false;
        if (handler != null) {
            preventDefault = handler.inputReceived(this, input);
        }

        // Default parsing and handling
        if (!preventDefault) {
            parseAndExecuteCommand(lastInput);
            prompt();
        }
    }

    private void parseAndExecuteCommand(String input) {
        String[] argv = parseInput(input);
        if (argv != null && argv.length > 0) {
            Command c = getCommand(argv[0]);
            if (c != null) {
                String result = executeCommand(c, argv);
                if (result != null) {
                    print(result);
                }
            } else {
                commandNotFound(argv);
            }
        }
    }

    protected void commandNotFound(String[] argv) {
        print("ERROR: " + argv[0] + ": command not found.");

    }

    protected String executeCommand(Command cmd, String[] argv) {
        try {
            Object r = cmd.execute(argv);
            return r != null ? "" + r : null;
        } catch (Exception e) {
            if (handler != null) {
                handler.exception(e, cmd, argv);
            } else {
                e.printStackTrace();
                println(e.getClass().getSimpleName() + ": " + e.getMessage());
            }
        }
        return null;
    }

    protected String[] parseInput(String input) {
        if (input != null && !"".equals(input.trim())) {
            String[] temp = input.split(" ");
            if (temp != null && temp.length > 0) {
                List<String> parsed = new ArrayList<String>(temp.length);
                String current = null;
                for (int i = 0; i < temp.length; i++) {
                    int quotCount = count(temp[i], '\"');
                    if (quotCount > 0 && quotCount % 2 != 0) {
                        // uneven number of quotes star or end combining params
                        if (current != null) {
                            parsed.add(current + " "
                                    + temp[i].replaceAll("\"", "")); // end
                            current = null;
                        } else {
                            current = temp[i].replaceAll("\"", ""); // start
                        }
                    } else if (current != null) {
                        current += " " + temp[i].replaceAll("\"", "");
                    } else {
                        parsed.add(temp[i].replaceAll("\"", ""));
                    }
                }

                // TODO: actually this is not quite right: We have an open quote
                // somewhere. Exception maybe?
                if (current != null) {
                    parsed.add(current.replaceAll("\"", ""));
                }
                return parsed.toArray(new String[] {});
            }
        }
        return new String[] {};
    }

    private int count(String sourceString, char lookFor) {
        if (sourceString == null) {
            return -1;
        }
        int count = 0;
        for (int i = 0; i < sourceString.length(); i++) {
            final char c = sourceString.charAt(i);
            if (c == lookFor) {
                count++;
            }
        }
        return count;
    }

    public void print(String output) {
        client.call("print", output);
    }

    public String getGreeting() {
        return config.greeting;
    }

    public String getPs() {
        return config.ps;
    }

    public int getMaxBufferSize() {
        return config.maxBufferSize;
    }

    public int getRows() {
        return config.rows;
    }

    public void setGreeting(String greeting) {
        config.greeting = greeting;
        client.call("setGreeting", greeting);
    }

    public void setPs(String ps) {
        config.ps = ps == null ? DEFAULT_PS : ps;
        client.call("setPs", config.ps);
    }

    public void setMaxBufferSize(int lines) {
        config.maxBufferSize = lines > 0 ? lines : 0;
        client.call("setMaxBufferSize", config.maxBufferSize);
    }

    public void setRows(int rows) {
        config.rows = rows;
        if (config.rows < 1) {
            config.rows = 1;
        }
        if (config.rows > MAX_ROWS) {
            config.rows = MAX_ROWS;
        }
        client.call("setRows", rows);
    }

    public int getCols() {
        return config.cols;
    }

    public void setCols(int cols) {
        config.cols = cols;
        if (config.cols < 1) {
            config.cols = 1;
        }
        if (config.cols > MAX_COLS) {
            config.cols = MAX_COLS;
        }
        client.call("setCols", config.cols);
    }

    public void prompt() {
        client.call("prompt");
    }

    public void prompt(String initialInput) {
        client.call("prompt", initialInput);
    }

    public void println(String string) {
        client.call("println", string);
    }

    public void newLine() {
        client.call("newLine");
    }

    public void reset() {
        client.call("reset");
    }

    public void clear() {
        formFeed();
    }

    public void formFeed() {
        client.call("ff");
    }

    public void carriageReturn() {
        client.call("cr");
    }

    public void lineFeed() {
        client.call("lf");
    }

    public void clearCommandHistory() {
        client.call("clearHistory");
    }

    public void clearBuffer() {
        client.call("clearBuffer");
    }

    public void scrollToEnd() {
        client.call("scrollToEnd");
    }

    /**
     * Gets the Tabulator index of this Focusable component.
     *
     * @see com.vaadin.ui.Component.Focusable#getTabIndex()
     */
    public int getTabIndex() {
        return tabIndex;
    }

    /**
     * Sets the Tabulator index of this Focusable component.
     *
     * @see com.vaadin.ui.Component.Focusable#setTabIndex(int)
     */
    public void setTabIndex(int tabIndex) {
        this.tabIndex = tabIndex;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void focus() {
        super.focus();
    }

    /* PrintStream implementation for console output. */

    public PrintStream getPrintStream() {
        if (printStream == null) {
            printStream = new PrintStream(new OutputStream() {

                ByteArrayOutputStream buffer = new ByteArrayOutputStream();

                @Override
                public void write(int b) throws IOException {
                    buffer.write(b);
                    // Line buffering
                    if (13 == b) {
                        flush();
                    }
                }

                @Override
                public void flush() throws IOException {
                    super.flush();
                    buffer.flush();
                    Console.this.print(buffer.toString());
                    buffer.reset();
                }
            }, true);
        }
        return printStream;
    }

    /* Generic command handling */

    public void addCommand(String name, Command cmd) {
        commands.put(name, cmd);
    }

    public void removeCommand(String cmdName) {
        commands.remove(cmdName);
    }

    public Command getCommand(String cmdName) {
        return commands.get(cmdName);
    }

    public Handler getHandler() {
        return handler;
    }

    public void setHandler(Handler handler) {
        this.handler = handler;
    }

    public Map<String, Command> getCommands() {
        return Collections.unmodifiableMap(commands);
    }

}
